import uuid

from opentrons.protocol_api._types import OffDeckType

from ..data.data import Data
from ..data.dbo.labware import LabwareWrapper


class LabwareLogic:
    def __init__(self, data: Data):
        self.data = data

    @staticmethod
    def labware_is_valid(instrument: LabwareWrapper | None):
        return instrument is not None and instrument.labware_instance is not None

    def child(self, labware_id: uuid.UUID):
        labware = self.get_valid_labware_by_id(labware_id)

        if labware.labware_instance.child is None:
            return None

        return self.data.static_items.get_generic_id_by_reference(labware.labware_instance.child)

    def parent(self, labware_id: uuid.UUID):
        labware = self.get_valid_labware_by_id(labware_id)

        parent = labware.labware_instance.parent

        if isinstance(parent, OffDeckType):
            return None

        if isinstance(parent, str):
            return parent


        return self.data.static_items.get_generic_id_by_reference(parent)

    def highest_z(self, labware_id: uuid.UUID):
        labware = self.get_valid_labware_by_id(labware_id)

        return labware.labware_instance.highest_z

    def tip_length(self, labware_id: uuid.UUID):
        labware = self.get_valid_labware_by_id(labware_id)

        return labware.labware_instance.tip_length

    def load_labware(self, labware_id: uuid.UUID, load_name: str, namespace: str = ""):
        parent_labware = self.get_valid_labware_by_id(labware_id)

        ns: str | None = namespace

        if len(namespace) == 0:
            ns = None

        child_labware = parent_labware.labware_instance.load_labware(name=load_name, namespace=ns)

        self.data.dynamic_items.register_wells_by_labware(labware=child_labware)

        return self.data.static_items.register_labware(labware=child_labware)

    def wells(self, labware_id: uuid.UUID):
        labware = self.get_valid_labware_by_id(labware_id)

        wells = labware.labware_instance.wells()

        well_ids = self.data.dynamic_items.get_well_ids_by_reference(wells)

        return well_ids

    def rows(self, labware_id: uuid.UUID):
        labware = self.get_valid_labware_by_id(labware_id)

        rows = labware.labware_instance.rows()

        new_rows = []

        for row in rows:
            new_rows.append(self.data.dynamic_items.get_well_ids_by_reference(row))

        return new_rows

    def columns(self, labware_id: uuid.UUID):
        labware = self.get_valid_labware_by_id(labware_id)

        columns = labware.labware_instance.columns()

        new_columns = []

        for column in columns:
            new_columns.append(self.data.dynamic_items.get_well_ids_by_reference(column))

        return new_columns

    def reset(self, labware_id: uuid.UUID):
        labware = self.get_valid_labware_by_id(labware_id)

        labware.labware_instance.reset()

    def get_valid_labware_by_id(self, labware_id: uuid.UUID):
        labware = self.data.static_items.get_labware_by_id(labware_id)
        if not self.labware_is_valid(labware):
            raise RuntimeError("Labware id not valid")
        return labware



# Generated by sila2.code_generator; sila2.__version__: 0.12.2
from .labware_base import LabwareBase
from .labware_client import LabwareClient
from .labware_feature import LabwareFeature
from .labware_types import DeleteLabware_Responses, MoveLabware_Responses, RegisterLabware_Responses

__all__ = [
    "LabwareBase",
    "LabwareFeature",
    "LabwareClient",
    "RegisterLabware_Responses",
    "DeleteLabware_Responses",
    "MoveLabware_Responses",
]



from opentrons.protocol_api import Labware
from opentrons.protocol_api._types import OffDeckType


def get_location(labware: Labware):
    parent = labware.parent
    while not isinstance(parent, str):
        if not isinstance(parent, OffDeckType):
            parent = parent.parent
        else:
            return None

    if not isinstance(parent, str):
        return None

    return parent


# Generated by sila2.code_generator; sila2.__version__: 0.12.2
from __future__ import annotations

from typing import NamedTuple


class RegisterLabware_Responses(NamedTuple):

    ItemIdentifier: str
    """
    UUID identifier for referenceable labware, modules,
                adapters, etc.
    """


class DeleteLabware_Responses(NamedTuple):

    pass


class MoveLabware_Responses(NamedTuple):

    pass



# Generated by sila2.code_generator; sila2.__version__: 0.12.2
from __future__ import annotations

from typing import NamedTuple


class Deactivate_Responses(NamedTuple):

    pass


class DeactivateBlockHeater_Responses(NamedTuple):

    pass


class DeactivateLidHeater_Responses(NamedTuple):

    pass


class CloseLid_Responses(NamedTuple):

    pass


class OpenLid_Responses(NamedTuple):

    pass


class SetLidTemperature_Responses(NamedTuple):

    pass


class SetBlockTemperature_Responses(NamedTuple):

    pass


class ExecuteProfile_Responses(NamedTuple):

    pass


class CurrentBlockTemperature_Responses(NamedTuple):

    Temperature: float
    """
    Temperature in Celsius
    """


class CurrentLidTemperature_Responses(NamedTuple):

    Temperature: float
    """
    Temperature in Celsius
    """


class CurrentLidTemperatureStatus_Responses(NamedTuple):

    TemperatureStatus: str
    """
    The temperature status
    """


class CurrentBlockTemperatureStatus_Responses(NamedTuple):

    TemperatureStatus: str
    """
    The temperature status
    """


class CurrentLidPosition_Responses(NamedTuple):

    LidPosition: str
    """
    The lid position
    """



import uuid
from typing import List

from opentrons.protocol_api import Well, TrashBin, WasteChute
from opentrons.types import Location

from ..data.data import Data

from ..data.dbo.module import ModuleWrapper
from helpers import get_location_from_str


class ModuleCommonLogic:
    def __init__(self, data: Data):
        self.data = data

    @staticmethod
    def module_is_valid(module: ModuleWrapper | None):
        return module is not None and module.module_instance is not None

    def verify_module(self, module_id):
        module = self.data.static_items.get_module_by_id(module_id)
        if not self.module_is_valid(module):
            raise RuntimeError("Module id not valid")
        return module

    def load_adapter(self, module_id: uuid.UUID, load_name: str, location: str, namespace: str | None):
        module = self.verify_module(module_id)

        old_adapter = self.data.static_items.get_labware_by_deck_location(module.module_instance.parent)
        if len(old_adapter) is not 0:
            for key, value in old_adapter.items():
                if value.data.load_name == load_name and value.data.namespace == namespace:
                    return key

        adapter = module.load_adapter(load_name=load_name,
                                                  location=get_location_from_str(location),
                                                  namespace=namespace)

        return self.data.static_items.register_labware(labware=adapter)

    def load_labware(self, module_id: uuid.UUID, load_name: str, location: str, namespace: str | None = None, adapter_name: str | None = None):
        module = self.verify_module(module_id)

        old_labware = self.data.static_items.get_labware_by_deck_location(module.module_instance.parent)
        if len(old_labware) is not 0:
            for key, value in old_labware.items():
                if value.data.load_name == load_name and value.data.namespace == namespace:
                    return key

        labware = module.load_labware(load_name=load_name,
                                                  location=get_location_from_str(location),
                                                  namespace=namespace,
                                                  adapter=adapter_name
                                                  )

        self.data.dynamic_items.register_wells_by_labware(labware=labware)

        return self.data.static_items.register_labware(labware=labware)

    def labware(self, module_id: uuid.UUID):
        module = self.verify_module(module_id)

        if module.module_instance.labware is None:
            return None

        return self.data.static_items.get_generic_id_by_reference(module.module_instance.labware)

    def parent(self, module_id: uuid.UUID):
        module = self.verify_module(module_id)

        if module.labware_instance is None:
            return None

        return module.module_instance.parent


# Generated by sila2.code_generator; sila2.__version__: 0.12.2
# -----
# This class does not do anything useful at runtime. Its only purpose is to provide type annotations.
# Since sphinx does not support .pyi files (yet?), this is a .py file.
# -----

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:

    from typing import Any, Iterable, List, Optional

    from modulethermocycler_types import (
        CloseLid_Responses,
        CurrentBlockTemperature_Responses,
        CurrentBlockTemperatureStatus_Responses,
        CurrentLidPosition_Responses,
        CurrentLidTemperature_Responses,
        CurrentLidTemperatureStatus_Responses,
        Deactivate_Responses,
        DeactivateBlockHeater_Responses,
        DeactivateLidHeater_Responses,
        ExecuteProfile_Responses,
        OpenLid_Responses,
        SetBlockTemperature_Responses,
        SetLidTemperature_Responses,
    )
    from sila2.client import ClientMetadataInstance, ClientObservableCommandInstance


class ModuleThermocyclerClient:
    """

    This feature contains all methods used for the thermocycler

    """

    def Deactivate(self, *, metadata: Optional[Iterable[ClientMetadataInstance]] = None) -> Deactivate_Responses:
        """
        Stops heating

        """
        ...

    def DeactivateBlockHeater(
        self, *, metadata: Optional[Iterable[ClientMetadataInstance]] = None
    ) -> DeactivateBlockHeater_Responses:
        """
        Stops heating

        """
        ...

    def DeactivateLidHeater(
        self, *, metadata: Optional[Iterable[ClientMetadataInstance]] = None
    ) -> DeactivateLidHeater_Responses:
        """
        Stops heating

        """
        ...

    def CloseLid(self, *, metadata: Optional[Iterable[ClientMetadataInstance]] = None) -> CloseLid_Responses:
        """
        Closes the absorbance plate lid

        """
        ...

    def OpenLid(self, *, metadata: Optional[Iterable[ClientMetadataInstance]] = None) -> OpenLid_Responses:
        """
        Opens the absorbance plate lid

        """
        ...

    def SetLidTemperature(
        self, Temperature: float, *, metadata: Optional[Iterable[ClientMetadataInstance]] = None
    ) -> SetLidTemperature_Responses:
        """

        Set the lid target temperature and waits for temperature to be reached.

        """
        ...

    def SetBlockTemperature(
        self,
        Temperature: float,
        HoldTimeMinutes: float,
        HoldTimeSeconds: float,
        BlockMaxVolume: float,
        *,
        metadata: Optional[Iterable[ClientMetadataInstance]] = None,
    ) -> SetBlockTemperature_Responses:
        """

        Set the block target temperature and waits for temperature to be reached.

        """
        ...

    def ExecuteProfile(
        self,
        Steps: List[Any],
        Repetitions: int,
        BlockMaxVolume: float,
        *,
        metadata: Optional[Iterable[ClientMetadataInstance]] = None,
    ) -> ExecuteProfile_Responses:
        """
        Executes the thermocycler profile
        """
        ...

    def CurrentBlockTemperature(
        self, *, metadata: Optional[Iterable[ClientMetadataInstance]] = None
    ) -> ClientObservableCommandInstance[CurrentBlockTemperature_Responses]:
        """
        Returns the current block temperature in Celsius

        """
        ...

    def CurrentLidTemperature(
        self, *, metadata: Optional[Iterable[ClientMetadataInstance]] = None
    ) -> ClientObservableCommandInstance[CurrentLidTemperature_Responses]:
        """
        Returns the current lid temperature in Celsius

        """
        ...

    def CurrentLidTemperatureStatus(
        self, *, metadata: Optional[Iterable[ClientMetadataInstance]] = None
    ) -> ClientObservableCommandInstance[CurrentLidTemperatureStatus_Responses]:
        """
        Returns the current lid temperature status

        """
        ...

    def CurrentBlockTemperatureStatus(
        self, *, metadata: Optional[Iterable[ClientMetadataInstance]] = None
    ) -> ClientObservableCommandInstance[CurrentBlockTemperatureStatus_Responses]:
        """
        Returns the current block temperature status

        """
        ...

    def CurrentLidPosition(
        self, *, metadata: Optional[Iterable[ClientMetadataInstance]] = None
    ) -> ClientObservableCommandInstance[CurrentLidPosition_Responses]:
        """
        Returns the current lid position
        """
        ...



# Generated by sila2.code_generator; sila2.__version__: 0.12.2
from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import timedelta
from typing import TYPE_CHECKING, Any, List, Optional

from sila2.server import FeatureImplementationBase, MetadataDict, ObservableCommandInstance

from .modulethermocycler_types import (
    CloseLid_Responses,
    CurrentBlockTemperature_Responses,
    CurrentBlockTemperatureStatus_Responses,
    CurrentLidPosition_Responses,
    CurrentLidTemperature_Responses,
    CurrentLidTemperatureStatus_Responses,
    Deactivate_Responses,
    DeactivateBlockHeater_Responses,
    DeactivateLidHeater_Responses,
    ExecuteProfile_Responses,
    OpenLid_Responses,
    SetBlockTemperature_Responses,
    SetLidTemperature_Responses,
)

if TYPE_CHECKING:

    from ...server import Server


class ModuleThermocyclerBase(FeatureImplementationBase, ABC):
    parent_server: Server

    CurrentBlockTemperature_default_lifetime_of_execution: Optional[timedelta]

    CurrentLidTemperature_default_lifetime_of_execution: Optional[timedelta]

    CurrentLidTemperatureStatus_default_lifetime_of_execution: Optional[timedelta]

    CurrentBlockTemperatureStatus_default_lifetime_of_execution: Optional[timedelta]

    CurrentLidPosition_default_lifetime_of_execution: Optional[timedelta]

    def __init__(self, parent_server: Server):
        """

        This feature contains all methods used for the thermocycler

        """
        super().__init__(parent_server=parent_server)

        self.CurrentBlockTemperature_default_lifetime_of_execution = None
        self.CurrentLidTemperature_default_lifetime_of_execution = None
        self.CurrentLidTemperatureStatus_default_lifetime_of_execution = None
        self.CurrentBlockTemperatureStatus_default_lifetime_of_execution = None
        self.CurrentLidPosition_default_lifetime_of_execution = None

    @abstractmethod
    def Deactivate(self, *, metadata: MetadataDict) -> Deactivate_Responses:
        """
        Stops heating



        :param metadata: The SiLA Client Metadata attached to the call

        """

    @abstractmethod
    def DeactivateBlockHeater(self, *, metadata: MetadataDict) -> DeactivateBlockHeater_Responses:
        """
        Stops heating



        :param metadata: The SiLA Client Metadata attached to the call

        """

    @abstractmethod
    def DeactivateLidHeater(self, *, metadata: MetadataDict) -> DeactivateLidHeater_Responses:
        """
        Stops heating



        :param metadata: The SiLA Client Metadata attached to the call

        """

    @abstractmethod
    def CloseLid(self, *, metadata: MetadataDict) -> CloseLid_Responses:
        """
        Closes the absorbance plate lid



        :param metadata: The SiLA Client Metadata attached to the call

        """

    @abstractmethod
    def OpenLid(self, *, metadata: MetadataDict) -> OpenLid_Responses:
        """
        Opens the absorbance plate lid



        :param metadata: The SiLA Client Metadata attached to the call

        """

    @abstractmethod
    def SetLidTemperature(self, Temperature: float, *, metadata: MetadataDict) -> SetLidTemperature_Responses:
        """

            Set the lid target temperature and waits for temperature to be reached.



        :param Temperature:
                A value between 37 and 110, representing the target temperature in °C.


        :param metadata: The SiLA Client Metadata attached to the call

        """

    @abstractmethod
    def SetBlockTemperature(
        self,
        Temperature: float,
        HoldTimeMinutes: float,
        HoldTimeSeconds: float,
        BlockMaxVolume: float,
        *,
        metadata: MetadataDict,
    ) -> SetBlockTemperature_Responses:
        """

            Set the block target temperature and waits for temperature to be reached.



        :param Temperature:
                A value between 4 and 99, representing the target temperature in °C.


        :param HoldTimeMinutes:
                The number of minutes to hold, after reaching temperature, before proceeding to the next command. Is added together with
                hold time seconds


        :param HoldTimeSeconds:
                The number of seconds to hold, after reaching temperature, before proceeding to the next command. Is added together with
                hold time minutes


        :param BlockMaxVolume:
                The greatest volume of liquid contained in any individual well of the loaded labware, in µL. If 0 is specified, the default is 25 µL.


        :param metadata: The SiLA Client Metadata attached to the call

        """

    @abstractmethod
    def ExecuteProfile(
        self, Steps: List[Any], Repetitions: int, BlockMaxVolume: float, *, metadata: MetadataDict
    ) -> ExecuteProfile_Responses:
        """
        Executes the thermocycler profile


        :param Steps: List of steps that make up a single cycle.

        :param Repetitions:
                The amount of times the steps will be repeated


        :param BlockMaxVolume:
                The greatest volume of liquid contained in any individual well of the loaded labware, in µL. If 0 is specified, the default is 25 µL.


        :param metadata: The SiLA Client Metadata attached to the call

        """

    @abstractmethod
    def CurrentBlockTemperature(
        self, *, metadata: MetadataDict, instance: ObservableCommandInstance
    ) -> CurrentBlockTemperature_Responses:
        """
        Returns the current block temperature in Celsius



        :param metadata: The SiLA Client Metadata attached to the call
        :param instance: The command instance, enabling sending status updates to subscribed clients

        :return:

            - Temperature: Temperature in Celsius


        """

    @abstractmethod
    def CurrentLidTemperature(
        self, *, metadata: MetadataDict, instance: ObservableCommandInstance
    ) -> CurrentLidTemperature_Responses:
        """
        Returns the current lid temperature in Celsius



        :param metadata: The SiLA Client Metadata attached to the call
        :param instance: The command instance, enabling sending status updates to subscribed clients

        :return:

            - Temperature: Temperature in Celsius


        """

    @abstractmethod
    def CurrentLidTemperatureStatus(
        self, *, metadata: MetadataDict, instance: ObservableCommandInstance
    ) -> CurrentLidTemperatureStatus_Responses:
        """
        Returns the current lid temperature status



        :param metadata: The SiLA Client Metadata attached to the call
        :param instance: The command instance, enabling sending status updates to subscribed clients

        :return:

            - TemperatureStatus: The temperature status


        """

    @abstractmethod
    def CurrentBlockTemperatureStatus(
        self, *, metadata: MetadataDict, instance: ObservableCommandInstance
    ) -> CurrentBlockTemperatureStatus_Responses:
        """
        Returns the current block temperature status



        :param metadata: The SiLA Client Metadata attached to the call
        :param instance: The command instance, enabling sending status updates to subscribed clients

        :return:

            - TemperatureStatus: The temperature status


        """

    @abstractmethod
    def CurrentLidPosition(
        self, *, metadata: MetadataDict, instance: ObservableCommandInstance
    ) -> CurrentLidPosition_Responses:
        """
        Returns the current lid position


        :param metadata: The SiLA Client Metadata attached to the call
        :param instance: The command instance, enabling sending status updates to subscribed clients

        :return:

            - LidPosition: The lid position


        """



import uuid
from dataclasses import dataclass
from typing import Literal, Dict

from opentrons.protocol_api import Labware, InstrumentContext, TrashBin, WasteChute, OFF_DECK
from opentrons.protocol_api._types import OffDeckType
from opentrons.protocol_api.core.labware import LabwareLoadParams, AbstractLabware
from opentrons.protocol_api.protocol_context import ModuleTypes
from opentrons.types import Point, DeckSlotName

from ..data import StaticItemsDict, DynamicItemsDict
from ..helpers import get_location


@dataclass
class LabwareData:
    load_name: str
    """Labware name used for loading the module"""
    location: str
    """Deck slot the labware or it's parent is in"""
    namespace: str | None = None
    """Namespace used for loading the labware"""
    parent: uuid.UUID | None = None
    """Item identifier of parent labware or module"""
    child: uuid.UUID | None = None
    """Item identifier of child labware"""

    def update(self, instance: Labware, static_items: StaticItemsDict):
        location = get_location(instance)
        if location is not None:
            self.location = location
        else:
            raise RuntimeError("Couldn't update location of labware: "+ self.load_name + " with location: " + self.location)

        self.child = static_items.get_generic_id_by_reference(instance.child)

        self.parent = static_items.get_generic_id_by_reference(instance.parent)

    @staticmethod
    def from_instance(instance: Labware, static_items: StaticItemsDict):
        namespace = instance.uri.split('/')[0]

        data = LabwareData(load_name=instance.load_name, location="", namespace=namespace)
        data.update(instance, static_items)
        return data


@dataclass
class LabwareWrapper:
    labware_instance: Labware | None
    """The instance used for usage in the Opentrons API"""
    data: LabwareData
    """Data needed to instantiate the instance"""

    def __eq__(self, other):
        if isinstance(other, LabwareWrapper):
            return self.labware_instance == other.labware_instance
        if isinstance(other, Labware):
            return self.labware_instance == other
        return False

    @staticmethod
    def from_instance(instance: Labware, static_items: StaticItemsDict):
        return LabwareWrapper(labware_instance=instance,
                              data=LabwareData.from_instance(instance=instance, static_items=static_items))



# Generated by sila2.code_generator; sila2.__version__: 0.12.2
from .modulethermocycler_base import ModuleThermocyclerBase
from .modulethermocycler_client import ModuleThermocyclerClient
from .modulethermocycler_feature import ModuleThermocyclerFeature
from .modulethermocycler_types import (
    CloseLid_Responses,
    CurrentBlockTemperature_Responses,
    CurrentBlockTemperatureStatus_Responses,
    CurrentLidPosition_Responses,
    CurrentLidTemperature_Responses,
    CurrentLidTemperatureStatus_Responses,
    Deactivate_Responses,
    DeactivateBlockHeater_Responses,
    DeactivateLidHeater_Responses,
    ExecuteProfile_Responses,
    OpenLid_Responses,
    SetBlockTemperature_Responses,
    SetLidTemperature_Responses,
)

__all__ = [
    "ModuleThermocyclerBase",
    "ModuleThermocyclerFeature",
    "ModuleThermocyclerClient",
    "Deactivate_Responses",
    "DeactivateBlockHeater_Responses",
    "DeactivateLidHeater_Responses",
    "CloseLid_Responses",
    "OpenLid_Responses",
    "SetLidTemperature_Responses",
    "SetBlockTemperature_Responses",
    "ExecuteProfile_Responses",
    "CurrentBlockTemperature_Responses",
    "CurrentLidTemperature_Responses",
    "CurrentLidTemperatureStatus_Responses",
    "CurrentBlockTemperatureStatus_Responses",
    "CurrentLidPosition_Responses",
]



# Generated by sila2.code_generator; sila2.__version__: 0.12.2
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from sila2.server import FeatureImplementationBase, MetadataDict

from .labware_types import DeleteLabware_Responses, MoveLabware_Responses, RegisterLabware_Responses

if TYPE_CHECKING:

    from ...server import Server


class LabwareBase(FeatureImplementationBase, ABC):
    parent_server: Server

    def __init__(self, parent_server: Server):
        """
        This feature allows you to manipulate the labware in the OT-2 and
        OT-Flex
        """
        super().__init__(parent_server=parent_server)

    @abstractmethod
    def RegisterLabware(
        self, SessionIdentifier: str, LoadName: str, Location: str, *, metadata: MetadataDict
    ) -> RegisterLabware_Responses:
        """
        This method registers labware to a specified position for
            the current session


        :param SessionIdentifier: UUID identifier for the current session

        :param LoadName: The name of the standard type of labware to register
                (https://labware.opentrons.com)

        :param Location: Location for labware, modules, adapters, etc.

        :param metadata: The SiLA Client Metadata attached to the call

        :return:

            - ItemIdentifier: UUID identifier for referenceable labware, modules,
                adapters, etc.


        """

    @abstractmethod
    def DeleteLabware(
        self, SessionIdentifier: str, Identifier: str, *, metadata: MetadataDict
    ) -> DeleteLabware_Responses:
        """
        This method deletes labware without removing it, Only use this
            method if you are removing it yourself and make sure to not execute
            any commands while moving it.


        :param SessionIdentifier: UUID identifier for the current session

        :param Identifier: Used to determine what labware to remove. This has to
                be an Item Identifier or a Location in the format used in other
                commands

        :param metadata: The SiLA Client Metadata attached to the call

        """

    @abstractmethod
    def MoveLabware(
        self, SessionIdentifier: str, ItemIdentifier: str, Location: str, *, metadata: MetadataDict
    ) -> MoveLabware_Responses:
        """
        This method deletes labware without removing it, Only use this
            method if you are removing it yourself and make sure to not execute
            any commands while moving it.


        :param SessionIdentifier: UUID identifier for the current session

        :param ItemIdentifier: UUID identifier for referencable labware, modules,
                adapters, etc.

        :param Location: Location for labware, modules, adapters, etc.

        :param metadata: The SiLA Client Metadata attached to the call

        """



import uuid
from dataclasses import dataclass
from typing import Literal, List

from opentrons.protocol_api import InstrumentContext

from ..data import StaticItemsDict, DynamicItemsDict


@dataclass
class InstrumentData:
    instrument_name: str
    """Instrument name used for loading the module"""
    mount: str | None = None
    """Where the instrument is mounted on the OT"""
    tip_racks: List[uuid.UUID] | None = None
    """Default tip racks assigned to this instrument"""

    def update(self, instance: InstrumentContext, static_items: StaticItemsDict):
        new_tip_racks = []

        for tip_rack in instance.tip_racks:
            tip_rack_id = static_items.get_generic_id_by_reference(tip_rack)
            if tip_rack_id is not None:
                new_tip_racks.append(tip_rack_id)
            else:
                raise RuntimeError()

        self.tip_racks = new_tip_racks

    @staticmethod
    def from_instance(instance: InstrumentContext, static_items: StaticItemsDict):
        mount: str | None = None
        if instance.channels != 96 and (instance.mount == "left" or instance.mount == "right"):
            mount = instance.mount

        data = InstrumentData(instrument_name=instance.name, mount=mount)
        data.update(instance, static_items)
        return data


@dataclass
class InstrumentWrapper:
    instrument_instance: InstrumentContext | None
    """The instance used for usage in the Opentrons API"""
    data: InstrumentData
    """Data needed to instantiate the instance"""

    def __eq__(self, other):
        if isinstance(other, InstrumentWrapper):
            return self.instrument_instance == other.instrument_instance
        if isinstance(other, InstrumentContext):
            return self.instrument_instance == other
        return False

    @staticmethod
    def from_instance(instance: InstrumentContext, static_items: StaticItemsDict):
        data = InstrumentData.from_instance(instance, static_items)
        data.update(instance, static_items)
        return InstrumentWrapper(instrument_instance=instance,
                                 data=data)



from opentrons.protocol_api import Labware, WasteChute, TrashBin
from opentrons.protocol_api.protocol_context import ModuleTypes
from ..data.dbo.labware import LabwareWrapper
from ..data.dbo.module import ModuleWrapper
from ..data.dbo.instrument import InstrumentWrapper
from ..data.dbo.trashbin import TrashBinWrapper


def get_location_from_str(string: str):
    if string.isnumeric():
        return int(string)
    else:
        return string

def convert_generic_to_labware_module_trash(generic:  LabwareWrapper | ModuleWrapper | InstrumentWrapper | TrashBinWrapper | WasteChute | None) -> Labware | ModuleTypes | WasteChute | TrashBin:
    if isinstance(generic, LabwareWrapper):
        return generic.labware_instance
    if isinstance(generic, ModuleWrapper):
        return generic.module_instance
    if isinstance(generic, TrashBinWrapper):
        return generic.trash_bin_instance
    if isinstance(generic, WasteChute):
        return generic
    raise RuntimeError("Id is not valid for this request")


import uuid
from dataclasses import dataclass
from typing import Literal

from opentrons.protocol_api import TrashBin
from opentrons.protocol_api.protocol_context import ModuleTypes
from opentrons.protocol_engine import ModuleModel
from opentrons.types import DeckSlotName
from opentrons_shared_data.module.types import ModuleType

from ..data import StaticItemsDict, DynamicItemsDict


@dataclass
class TrashBinData:
    deck_location: str
    """Location on the OT deck the module is loaded on"""
    is_fixed: bool

@dataclass
class TrashBinWrapper:
    trash_bin_instance: TrashBin | None
    """The instance used for usage in the Opentrons API"""
    data: TrashBinData
    """Data needed to instantiate the instance"""

    def __eq__(self, other):
        if isinstance(other, TrashBinWrapper):
            return self.trash_bin_instance == other.trash_bin_instance
        if isinstance(other, TrashBin):
            return self.trash_bin_instance == other
        return False


import uuid
from typing import List

from opentrons.protocol_api import Well, TrashBin, WasteChute
from opentrons.types import Location

from ..data.data import Data

from ..data.dbo.instrument import InstrumentWrapper


class InstrumentLogic:
    def __init__(self, data: Data):
        self.data = data

    @staticmethod
    def instrument_is_valid(instrument: InstrumentWrapper | None):
        return instrument is not None and instrument.instrument_instance is not None

    def home_plunger(self, instrument_id: uuid.UUID):
        instrument = self.data.static_items.get_instrument_by_id(instrument_id)

        if not self.instrument_is_valid(instrument):
            raise RuntimeError("Instrument id not valid")

        instrument.instrument_instance.home_plunger()

    def configure_for_volume(self, instrument_id: uuid.UUID, volume: float):
        instrument = self.data.static_items.get_instrument_by_id(instrument_id)

        if not self.instrument_is_valid(instrument):
            raise RuntimeError("Instrument id not valid")

        instrument.instrument_instance.configure_for_volume(volume=volume)

    def set_tip_racks(self, instrument_id: uuid.UUID, tip_rack_ids: List[uuid.UUID]):
        instrument = self.data.static_items.get_instrument_by_id(instrument_id)

        if not self.instrument_is_valid(instrument):
            raise RuntimeError("Instrument id not valid")

        tip_racks = []
        for tip_rack_id in tip_rack_ids:
            tip_rack = self.data.static_items.get_labware_by_id(tip_rack_id)
            if tip_rack is None or tip_rack.labware_instance is None or not tip_rack.labware_instance.is_tiprack:
                raise RuntimeError("Tip rack id not valid")
            tip_racks.append(tip_rack.labware_instance)

        instrument.data.tip_racks = tip_rack_ids
        instrument.instrument_instance.tip_racks = tip_racks

    def drop_tip(self, instrument_id: uuid.UUID, home_after: bool = False):
        instrument = self.data.static_items.get_instrument_by_id(instrument_id)

        if not self.instrument_is_valid(instrument):
            raise RuntimeError("Instrument id not valid")

        instrument.instrument_instance.drop_tip(home_after=home_after)

    def drop_tip_location(self, instrument_id: uuid.UUID, location_id: uuid.UUID, home_after: bool = False):
        instrument = self.data.static_items.get_instrument_by_id(instrument_id)

        if not self.instrument_is_valid(instrument):
            raise RuntimeError("Instrument id not valid")

        location = self.data.get_well_location_trash_by_id(location_id)

        if location is None or isinstance(location, Well):
            raise RuntimeError("Location or Trash id not valid")

        instrument.instrument_instance.drop_tip(home_after=home_after)

    def pick_up_tip(self, instrument_id: uuid.UUID):
        instrument = self.data.static_items.get_instrument_by_id(instrument_id)

        if not self.instrument_is_valid(instrument):
            raise RuntimeError("Instrument id not valid")

        instrument.instrument_instance.pick_up_tip()

    def return_tip(self, instrument_id: uuid.UUID):
        instrument = self.data.static_items.get_instrument_by_id(instrument_id)

        if not self.instrument_is_valid(instrument):
            raise RuntimeError("Instrument id not valid")

        instrument.instrument_instance.return_tip()

    @staticmethod
    def prepare_air_gap_parameters(volume, height):
        volume_value: float | None = volume
        height_value: float | None = height
        if volume_value == 0:
            volume_value = None
        if height_value == 0:
            height_value = None
        return height_value, volume_value

    def air_gap(self, instrument_id: uuid.UUID, volume: float = 0, height: float = 0):
        instrument = self.data.static_items.get_instrument_by_id(instrument_id)

        if not self.instrument_is_valid(instrument):
            raise RuntimeError("Instrument id not valid")

        height_value, volume_value = self.prepare_air_gap_parameters(volume, height)

        instrument.instrument_instance.air_gap(volume=volume_value, height=height_value)

    @staticmethod
    def prepare_aspirate_parameters(rate, volume):
        volume_value: float | None = volume
        rate_value: float | None = rate
        if volume_value == 0:
            volume_value = None
        if rate_value == 0:
            rate_value = None
        return rate_value, volume_value

    def aspirate(self, instrument_id: uuid.UUID, volume: float = 0, rate: float = 0):
        instrument = self.data.static_items.get_instrument_by_id(instrument_id)

        if not self.instrument_is_valid(instrument):
            raise RuntimeError("Instrument id not valid")

        rate_value, volume_value = self.prepare_aspirate_parameters(volume, rate)

        instrument.instrument_instance.aspirate(volume=volume_value, rate=rate_value)

    def aspirate_location(self, instrument_id: uuid.UUID, location_id: uuid.UUID, volume: float = 0, rate: float = 0):
        instrument = self.data.static_items.get_instrument_by_id(instrument_id)

        if not self.instrument_is_valid(instrument):
            raise RuntimeError("Instrument id not valid")

        rate_value, volume_value = self.prepare_aspirate_parameters(volume, rate)

        location = self.data.dynamic_items.get_well_location_by_id(location_id)

        if location is None:
            raise RuntimeError("Well location id not valid")

        instrument.instrument_instance.aspirate(volume=volume_value, location=location, rate=rate_value)

    def blow_out(self, instrument_id: uuid.UUID):
        instrument = self.data.static_items.get_instrument_by_id(instrument_id)

        if not self.instrument_is_valid(instrument):
            raise RuntimeError("Instrument id not valid")

        instrument.instrument_instance.blow_out()

    def blow_out_location(self, instrument_id: uuid.UUID, location_id: uuid.UUID):
        instrument = self.data.static_items.get_instrument_by_id(instrument_id)

        if not self.instrument_is_valid(instrument):
            raise RuntimeError("Instrument id not valid")

        location = self.data.get_well_location_trash_by_id(location_id)

        if location is None:
            raise RuntimeError("Well, Location or Trash id not valid")

        instrument.instrument_instance.blow_out(location=location)

    @staticmethod
    def prepare_mix_variables(rate, repetitions, volume):
        repetitions_value: int = repetitions
        volume_value: float | None = volume
        rate_value: float | None = rate
        if repetitions_value == 0:
            repetitions_value = 1
        if volume_value == 0:
            volume_value = None
        if rate_value == 0:
            rate_value = None
        return rate_value, repetitions_value, volume_value

    def mix(self, instrument_id: uuid.UUID, location_id: uuid.UUID, repetitions: int = 0, volume: float = 0, rate: float = 0):
        instrument = self.data.static_items.get_instrument_by_id(instrument_id)

        rate_value, repetitions_value, volume_value = self.prepare_mix_variables(rate, repetitions, volume)

        if not self.instrument_is_valid(instrument):
            raise RuntimeError("Instrument id not valid")

        instrument.instrument_instance.mix(repetitions=repetitions_value, volume=volume_value, rate=rate_value)

    def mix_location(self, instrument_id: uuid.UUID, location_id: uuid.UUID, repetitions: int = 0, volume: float = 0, rate: float = 0):
        instrument = self.data.static_items.get_instrument_by_id(instrument_id)

        rate_value, repetitions_value, volume_value = self.prepare_mix_variables(rate, repetitions, volume)

        if not self.instrument_is_valid(instrument):
            raise RuntimeError("Instrument id not valid")

        location = self.data.dynamic_items.get_well_location_by_id(location_id)

        if location is None:
            raise RuntimeError("Well location id not valid")

        instrument.instrument_instance.mix(location=location, repetitions=repetitions_value, volume=volume_value, rate=rate_value)

    @staticmethod
    def prepare_dispense_parameters(push_out, rate, volume):
        volume_value: float | None = volume
        rate_value: float | None = rate
        push_out_value: float | None = push_out
        if volume_value == 0:
            volume_value = None
        if rate_value == 0:
            rate_value = None
        if push_out_value == 0:
            push_out_value = None
        return push_out_value, rate_value, volume_value

    def dispense(self, instrument_id: uuid.UUID, volume: float = 0, rate: float = 0, push_out: float = 0):
        instrument = self.data.static_items.get_instrument_by_id(instrument_id)

        push_out_value, rate_value, volume_value = self.prepare_dispense_parameters(push_out, rate, volume)

        if not self.instrument_is_valid(instrument):
            raise RuntimeError("Instrument id not valid")

        instrument.instrument_instance.dispense(volume=volume_value, rate=rate_value, push_out=push_out_value)

    def dispense_location(self, instrument_id: uuid.UUID, location_id: uuid.UUID, volume: float = 0, rate: float = 0, push_out: float = 0):
        instrument = self.data.static_items.get_instrument_by_id(instrument_id)

        push_out_value, rate_value, volume_value = self.prepare_dispense_parameters(push_out, rate, volume)

        if not self.instrument_is_valid(instrument):
            raise RuntimeError("Instrument id not valid")

        location = self.data.get_well_location_trash_by_id(location_id)

        if location is None:
            raise RuntimeError("Well, Location or Trash id not valid")

        instrument.instrument_instance.dispense(volume=volume_value, location=location, rate=rate_value,
                                                push_out=push_out_value)

    def move_to_location(self, instrument_id: uuid.UUID, location_id: uuid.UUID, force_direct: bool = False, speed: float = 0):
        instrument = self.data.static_items.get_instrument_by_id(instrument_id)

        if speed == 0:
            speed_value = None
        else:
            speed_value = speed

        if not self.instrument_is_valid(instrument):
            raise RuntimeError("Instrument id not valid")

        location = self.data.get_well_location_trash_by_id(location_id)

        if location is None or isinstance(location, Well):
            raise RuntimeError("Location or Trash id not valid")

        instrument.instrument_instance.move_to(location=location, force_direct=force_direct, speed=speed_value)


# Generated by sila2.code_generator; sila2.__version__: 0.12.2
from __future__ import annotations

from datetime import timedelta
from typing import TYPE_CHECKING, Any, List

from sila2.server import MetadataDict, ObservableCommandInstance

from ..generated.modulethermocycler import (
    CloseLid_Responses,
    CurrentBlockTemperature_Responses,
    CurrentBlockTemperatureStatus_Responses,
    CurrentLidPosition_Responses,
    CurrentLidTemperature_Responses,
    CurrentLidTemperatureStatus_Responses,
    Deactivate_Responses,
    DeactivateBlockHeater_Responses,
    DeactivateLidHeater_Responses,
    ExecuteProfile_Responses,
    ModuleThermocyclerBase,
    OpenLid_Responses,
    SetBlockTemperature_Responses,
    SetLidTemperature_Responses,
)

if TYPE_CHECKING:
    from ..server import Server


class ModuleThermocyclerImpl(ModuleThermocyclerBase):
    def __init__(self, parent_server: Server) -> None:
        super().__init__(parent_server=parent_server)

        # Default lifetime of observable command instances. Possible values:
        # None: Command instance is valid and stored in memory until server shutdown
        # datetime.timedelta: Command instance is deleted after this duration, can be increased during command runtime
        self.CurrentBlockTemperature_default_lifetime_of_execution = timedelta(minutes=30)
        self.CurrentLidTemperature_default_lifetime_of_execution = timedelta(minutes=30)
        self.CurrentLidTemperatureStatus_default_lifetime_of_execution = timedelta(minutes=30)
        self.CurrentBlockTemperatureStatus_default_lifetime_of_execution = timedelta(minutes=30)
        self.CurrentLidPosition_default_lifetime_of_execution = timedelta(minutes=30)

    def Deactivate(self, *, metadata: MetadataDict) -> Deactivate_Responses:
        raise NotImplementedError  # TODO

    def DeactivateBlockHeater(self, *, metadata: MetadataDict) -> DeactivateBlockHeater_Responses:
        raise NotImplementedError  # TODO

    def DeactivateLidHeater(self, *, metadata: MetadataDict) -> DeactivateLidHeater_Responses:
        raise NotImplementedError  # TODO

    def CloseLid(self, *, metadata: MetadataDict) -> CloseLid_Responses:
        raise NotImplementedError  # TODO

    def OpenLid(self, *, metadata: MetadataDict) -> OpenLid_Responses:
        raise NotImplementedError  # TODO

    def SetLidTemperature(self, Temperature: float, *, metadata: MetadataDict) -> SetLidTemperature_Responses:
        raise NotImplementedError  # TODO

    def SetBlockTemperature(
        self,
        Temperature: float,
        HoldTimeMinutes: float,
        HoldTimeSeconds: float,
        BlockMaxVolume: float,
        *,
        metadata: MetadataDict,
    ) -> SetBlockTemperature_Responses:
        raise NotImplementedError  # TODO

    def ExecuteProfile(
        self, Steps: List[Any], Repetitions: int, BlockMaxVolume: float, *, metadata: MetadataDict
    ) -> ExecuteProfile_Responses:
        raise NotImplementedError  # TODO

    def CurrentBlockTemperature(
        self, *, metadata: MetadataDict, instance: ObservableCommandInstance
    ) -> CurrentBlockTemperature_Responses:
        # set execution status from `waiting` to `running`
        instance.begin_execution()

        raise NotImplementedError  # TODO

    def CurrentLidTemperature(
        self, *, metadata: MetadataDict, instance: ObservableCommandInstance
    ) -> CurrentLidTemperature_Responses:
        # set execution status from `waiting` to `running`
        instance.begin_execution()

        raise NotImplementedError  # TODO

    def CurrentLidTemperatureStatus(
        self, *, metadata: MetadataDict, instance: ObservableCommandInstance
    ) -> CurrentLidTemperatureStatus_Responses:
        # set execution status from `waiting` to `running`
        instance.begin_execution()

        raise NotImplementedError  # TODO

    def CurrentBlockTemperatureStatus(
        self, *, metadata: MetadataDict, instance: ObservableCommandInstance
    ) -> CurrentBlockTemperatureStatus_Responses:
        # set execution status from `waiting` to `running`
        instance.begin_execution()

        raise NotImplementedError  # TODO

    def CurrentLidPosition(
        self, *, metadata: MetadataDict, instance: ObservableCommandInstance
    ) -> CurrentLidPosition_Responses:
        # set execution status from `waiting` to `running`
        instance.begin_execution()

        raise NotImplementedError  # TODO



import uuid
from typing import List

from opentrons.motion_planning.deck_conflict import HeaterShakerModule
from opentrons.protocol_api import Well, TrashBin, WasteChute
from opentrons.types import Location

from ..data.data import Data

from ..data.dbo.module import ModuleWrapper
from helpers import get_location_from_str


class ModuleHeaterShakerLogic:
    def __init__(self, data: Data):
        self.data = data

    @staticmethod
    def module_is_valid(module: ModuleWrapper | None):
        return module is not None and module.module_instance is not None

    def get_valid_module_by_id(self, module_id):
        module: ModuleWrapper = self.data.static_items.get_module_by_id(module_id)
        if not self.module_is_valid(module):
            raise RuntimeError("Module Heater Shaker id not valid")
        return module

    def close_labware_latch(self, module_id: uuid.UUID):
        module = self.get_valid_module_by_id(module_id)

        module.module_instance.close_labware_latch()

    def open_labware_latch(self, module_id: uuid.UUID):
        module = self.get_valid_module_by_id(module_id)

        module.module_instance.open_labware_latch()

    def deactivate_heater(self, module_id: uuid.UUID):
        module = self.get_valid_module_by_id(module_id)

        module.module_instance.deactivate_heater()

    def deactivate_shaker(self, module_id: uuid.UUID):
        module = self.get_valid_module_by_id(module_id)

        module.module_instance.deactivate_shaker()

    def set_and_wait_for_shake_speed(self, module_id: uuid.UUID, rpm: int):
        module = self.get_valid_module_by_id(module_id)

        module.module_instance.set_and_wait_for_shake_speed(rpm)

    def set_and_wait_for_temperature(self, module_id: uuid.UUID, celsius: float):
        module = self.get_valid_module_by_id(module_id)

        module.module_instance.set_and_wait_for_temperature(celsius)

    def set_target_temperature(self, module_id: uuid.UUID, celsius: float):
        module = self.get_valid_module_by_id(module_id)

        module.module_instance.set_target_temperature(celsius)

    def wait_for_temperature(self, module_id: uuid.UUID):
        module = self.get_valid_module_by_id(module_id)

        module.module_instance.wait_for_temperature()
    def load_adapter(self, module_id: uuid.UUID, load_name: str, location: str, namespace: str | None):
        module: ModuleWrapper = self.data.static_items.get_module_by_id(module_id)

        if not self.module_is_valid(module):
            raise RuntimeError("Module Heater Shaker id not valid")

        old_adapter = self.data.static_items.get_labware_by_deck_location(location)
        if len(old_adapter) is not 0:
            for key, value in old_adapter.items():
                if value.data.load_name == load_name and value.data.namespace == namespace:
                    return key

        adapter = module.module_instance.load_adapter(name=load_name,
                                                      namespace=namespace)

        return self.data.static_items.register_labware(labware=adapter)

    def load_labware(self, module_id: uuid.UUID, load_name: str, location: str, namespace: str | None = None, adapter_name: str | None = None):
        module: ModuleWrapper = self.data.static_items.get_module_by_id(module_id)

        if not self.module_is_valid(module):
            raise RuntimeError("Module Heater Shaker id not valid")

        old_labware = self.data.static_items.get_labware_by_deck_location(location)
        if len(old_labware) is not 0:
            for key, value in old_labware.items():
                if value.data.load_name == load_name and value.data.namespace == namespace:
                    return key

        labware = module.module_instance.load_labware(name=load_name,
                                                      namespace=namespace,
                                                      adapter=adapter_name)

        self.data.dynamic_items.register_wells_by_labware(labware=labware)

        return self.data.static_items.register_labware(labware=labware)


import uuid
from dataclasses import dataclass
from typing import Literal

from opentrons.protocol_api.protocol_context import ModuleTypes
from opentrons.protocol_engine import ModuleModel
from opentrons.types import DeckSlotName
from opentrons_shared_data.module.types import ModuleType

from ..data import StaticItemsDict, DynamicItemsDict


@dataclass
class ModuleData:
    module_name: str
    """Module Type used for loading the module"""
    location: str
    """Location on the OT deck the module is loaded on"""
    child: uuid.UUID | None = None
    """Item identifier of child labware if applicable"""

    def update(self, instance: ModuleTypes, static_items: StaticItemsDict):
        self.child = static_items.get_generic_id_by_reference(instance.labware)

    @staticmethod
    def from_instance(instance: ModuleTypes, static_items: StaticItemsDict):
        data = ModuleData(module_name=instance.model, location=instance.parent, )
        data.update(instance, static_items)
        return data


@dataclass
class ModuleWrapper:
    module_instance: ModuleTypes | None
    """The instance used for usage in the Opentrons API"""
    data: ModuleData
    """Data needed to instantiate the instance"""

    def __eq__(self, other):
        if isinstance(other, ModuleWrapper):
            return self.module_instance == other.module_instance
        if isinstance(other, ModuleTypes):
            return self.module_instance == other
        return False

    @staticmethod
    def from_instance(instance: ModuleTypes, static_items: StaticItemsDict):
        data = ModuleData.from_instance(instance, static_items)
        data.update(instance, static_items)
        return ModuleWrapper(module_instance=instance,
                             data=data)



import uuid
from typing import Literal, List

from opentrons.protocol_api import WasteChute
from opentrons.protocol_api._types import OffDeckType
from opentrons.protocol_api.core.labware import LabwareLoadParams
from opentrons.types import DeckSlotName

from ..data.data import Data, waste_chute_id, fixed_trash_id

from .helpers import get_location_from_str, convert_generic_to_labware_module_trash


class ProtocolLogic:
    def __init__(self, data: Data):
        self.data = data

    def set_rail_lights(self, on: bool):
        self.data.protocol.set_rail_lights(on)

    @property
    def door_closed(self):
        return self.data.protocol.door_closed

    def clear_commands(self):
        self.data.protocol.clear_commands()

    def commands(self):
        return self.data.protocol.commands()

    def home(self):
        instruments = [self.data.static_items.get_instrument_by_mount("left"),
                       self.data.static_items.get_instrument_by_mount("right"),
                       self.data.static_items.get_instrument_by_mount(None)]

        for instrument in instruments:
            if instrument is not None:
                instrument.instrument_instance.home_plunger()

        return self.data.protocol.home()

    def load_fixed_trash_bin(self):
        if self.data.static_items.has_fixed_trash:
            return fixed_trash_id

        fixed_trash = self.data.protocol.fixed_trash
        return self.data.static_items.register_trash(trash=fixed_trash, is_fixed=True)

    def load_waste_chute(self):
        if self.data.static_items.has_waste_chute:
            return waste_chute_id
        waste_chute = self.data.protocol.load_waste_chute()
        return self.data.static_items.register_trash(trash=waste_chute, is_waste_chute=True)

    def load_trash_bin(self, location: str):
        old_trash_bin = self.data.static_items.get_trash_id_by_deck_location(location)
        if old_trash_bin is not None:
            return old_trash_bin

        trash_bin = self.data.protocol.load_trash_bin(location=get_location_from_str(location))
        return self.data.static_items.register_trash(trash=trash_bin)

    def load_module(self, module_name: str, location: str):
        old_module = self.data.static_items.get_module_by_deck_location(location)
        if old_module is not None and old_module.data.module_name == module_name:
            return self.data.static_items.get_module_id_by_reference(old_module.module_instance)

        module = self.data.protocol.load_module(module_name=module_name,
                                                location=get_location_from_str(location))
        return self.data.static_items.register_module(module=module)

    def load_adapter(self, load_name: str, location: str, namespace: str | None):
        old_adapter = self.data.static_items.get_labware_by_deck_location(location)
        if len(old_adapter) is not 0:
            for key, value in old_adapter.items():
                if value.data.load_name == load_name and value.data.namespace == namespace:
                    return key

        adapter = self.data.protocol.load_adapter(load_name=load_name,
                                                  location=get_location_from_str(location),
                                                  namespace=namespace)

        return self.data.static_items.register_labware(labware=adapter)

    def load_labware(self, load_name: str, location: str, namespace: str | None = None, adapter_name: str | None = None):
        old_labware = self.data.static_items.get_labware_by_deck_location(location)
        if len(old_labware) is not 0:
            for key, value in old_labware.items():
                if value.data.load_name == load_name and value.data.namespace == namespace:
                    return key

        labware = self.data.protocol.load_labware(load_name=load_name,
                                                  location=get_location_from_str(location),
                                                  namespace=namespace,
                                                  adapter=adapter_name
                                                  )

        self.data.dynamic_items.register_wells_by_labware(labware=labware)

        return self.data.static_items.register_labware(labware=labware)

    def load_instrument(self, instrument_name: str, mount: str, tip_racks: List[uuid.UUID]):
        old_instrument = self.data.static_items.get_instrument_by_mount(mount)
        if old_instrument is not None:
            return self.data.static_items.get_instrument_id_by_reference(old_instrument.instrument_instance)

        tip_rack_items = []
        for tip_rack in tip_racks:
            item = self.data.static_items.get_labware_by_id(tip_rack)
            if item is not None:
                tip_rack_items.append(item)

        instrument = self.data.protocol.load_instrument(instrument_name=instrument_name,
                                                        mount=mount,
                                                        tip_racks=tip_rack_items,
                                                        liquid_presence_detection=True)
        return self.data.static_items.register_instrument(instrument=instrument)

    def move_labware(self, labware_id: uuid.UUID, new_location: str,
                     pick_up_x: float = 0, pick_up_y: float = 0, pick_up_z: float = 0,
                     drop_x: float = 0, drop_y: float = 0, drop_z: float = 0):
        labware = self.data.static_items.get_labware_by_id(labware_id)
        self.data.protocol.move_labware(labware=labware.labware_instance,
                                        new_location=get_location_from_str(new_location),
                                        pick_up_offset={"x": pick_up_x, "y": pick_up_y, "z": pick_up_z},
                                        drop_offset={"x": drop_x, "y": drop_y, "z": drop_z},
                                        use_gripper=True)

    def move_labware_to_id(self, labware_id: uuid.UUID, new_location_id: uuid.UUID,
                           pick_up_x: float = 0, pick_up_y: float = 0, pick_up_z: float = 0,
                           drop_x: float = 0, drop_y: float = 0, drop_z: float = 0):
        labware = self.data.static_items.get_labware_by_id(labware_id)
        new_location = self.data.static_items.get_generic_by_id(new_location_id)

        if new_location is None:
            raise RuntimeError("Location id invalid")

        self.data.protocol.move_labware(labware=labware.labware_instance,
                                        new_location=convert_generic_to_labware_module_trash(new_location),
                                        pick_up_offset={"x": pick_up_x, "y": pick_up_y, "z": pick_up_z},
                                        drop_offset={"x": drop_x, "y": drop_y, "z": drop_z},
                                        use_gripper=True)



import uuid
from typing import Dict, Literal, Protocol, List

from blackd.middlewares import middleware
from opentrons.protocol_api import Labware, ProtocolContext, InstrumentContext, TrashBin, WasteChute, Well
from opentrons.protocol_api.protocol_context import ModuleTypes
from opentrons.types import DeckSlotName, Location

from dbo.instrument import InstrumentWrapper, InstrumentData
from dbo.labware import LabwareWrapper, LabwareData
from dbo.module import ModuleWrapper, ModuleData
from dbo.trashbin import TrashBinWrapper, TrashBinData

fixed_trash_id = uuid.UUID("8c1004ae-a3b0-4230-a3d1-cf5043d7088c")
waste_chute_id = uuid.UUID("e4114551-24cb-4e5e-9068-38422be95b5b")


class Data:
    def __init__(self, protocol: ProtocolContext, is_ot_flex=True):
        self.protocol: ProtocolContext = protocol
        self.static_items: StaticItemsDict = StaticItemsDict(self)
        self.dynamic_items: DynamicItemsDict = DynamicItemsDict(self)
        self.is_ot_flex: bool = is_ot_flex

    def get_well_location_trash_by_id(self, item_id: uuid.UUID):
        location: Well | Location | TrashBin | WasteChute | None = self.dynamic_items.get_well_location_by_id(item_id)
        trash: TrashBinWrapper | WasteChute | None = None

        if location is None:
            trash = self.static_items.get_trash_by_id(item_id)

        if trash is None:
            return None

        if isinstance(trash, TrashBinWrapper):
            location = trash.trash_bin_instance
        else:
            location = trash

        return location

    def generate_new_id(self):
        generated_uuid = uuid.uuid4()
        while (self.static_items.dict.get(generated_uuid) is not None
               or self.dynamic_items.dict.get(generated_uuid) is not None
               or fixed_trash_id == generated_uuid
               or waste_chute_id == generated_uuid):
            generated_uuid = uuid.uuid4()
        return generated_uuid


class StaticItemsDict:
    dict: Dict[uuid.UUID, LabwareWrapper | ModuleWrapper | InstrumentWrapper | TrashBinWrapper | WasteChute] = {}
    data: Data

    def __init__(self, data: Data):
        self.has_fixed_trash: bool = False
        self.has_waste_chute: bool = False
        self.data = data

    # def apply(self, protocol: ProtocolContext):
    #     if self.has_fixed_trash:
    #         self.register_trash(self.get_fixed_trash_id(), protocol.fixed_trash)
    #
    #     if self.has_waste_chute:
    #        self.register_trash(self.get_waste_chute_id(), protocol.load_waste_chute())
    #
    #     for key, value in self.dict.items():
    #         match value:
    #             case WasteChuteWrapper():
    #                 value.
    #
    #
    #         if isinstance(value, WasteChuteWrapper):
    #             value.waste_chute_instance = protocol.load_waste_chute()
    #         else:
    #             if isinstance(value, TrashBinWrapper):
    #                 if value.data.is_fixed:
    #                     value.trash_bin_instance = protocol.fixed_trash
    #                 else:
    #                     value.trash_bin_instance = protocol.load_trash_bin(str(value.data.deck_location))
    #             else:
    #                 if isinstance(value, ModuleWrapper):
    #                     value.module_instance
    #
    #
    #
    #     """Applies all the objects to the protocol given"""

    def get_generic_by_id(self, item_id: uuid.UUID):
        return self.dict.get(item_id)

    def get_generic_id_by_reference(self, reference: Labware | ModuleTypes | InstrumentContext | TrashBin | WasteChute | None):
        if reference == None:
            return None
        item = None
        for key, value in self.dict.items():
            if value == reference:
                item = key
        return item

    def update_generic_by_id(self, item_id: uuid.UUID):
        item = self.get_generic_by_id(item_id)
        self.update_generic(item)

    def update_generic(self, item: LabwareWrapper | ModuleWrapper | InstrumentWrapper | TrashBinWrapper | WasteChute):
        if isinstance(item, LabwareWrapper):
            item.data.update(item.labware_instance, self)
            return
        if isinstance(item, ModuleWrapper):
            item.data.update(item.module_instance, self)
            return
        if isinstance(item, InstrumentWrapper):
            item.data.update(item.instrument_instance, self)
            return

    def update_all(self):
        for key, item in self.dict.items():
            self.update_generic(item)

    def register_labware(self, labware: Labware):
        wrapper = LabwareWrapper.from_instance(labware, self)
        child = wrapper.labware_instance.child
        if child is not None:
            self.register_labware(child)

        new_uuid = self.generate_new_id()
        self.dict[new_uuid] = wrapper
        return new_uuid

    def unregister_labware(self, item_id: uuid.UUID):
        labware = self.get_labware_by_id(item_id)

        to_delete = [item_id]

        if labware is None:
            return

        if labware is not None:
            child = labware.data.child
            while child is not None:
                to_delete.append(child)
                child_instance = self.get_labware_by_id(child)
                child = child_instance.data.child

        for to_delete_id in to_delete:
            labware = self.get_labware_by_id(to_delete_id)
            if labware is not None:
                self.data.dynamic_items.unregister_wells(labware.labware_instance)
            self.dict.pop(to_delete_id)

    def get_labware_by_id(self, item_id: uuid.UUID):
        item = self.dict.get(item_id)

        if isinstance(item, LabwareWrapper):
            return item

        return None

    def get_labware_id_by_reference(self, labware: Labware):
        item = None
        for key, value in self.dict.items():
            if value == labware:
                return key
        return item

    def get_labware_by_deck_location(self, location: str):
        items = {}

        for key, value in self.dict.items():
            if isinstance(value, LabwareWrapper) and value.data.location == location:
                items[key] = value

        return items

    def register_module(self, module: ModuleTypes):
        wrapper = ModuleWrapper.from_instance(module, self)
        new_uuid = self.generate_new_id()
        self.dict[new_uuid] = wrapper
        return new_uuid

    def get_module_by_id(self, item_id: uuid.UUID):
        item = self.dict.get(item_id)

        if isinstance(item, ModuleWrapper):
            return item

        return None

    def get_module_id_by_reference(self, module: ModuleTypes):
        item = None
        for key, value in self.dict.items():
            if value == module:
                item = key

        return item

    def get_module_by_deck_location(self, location: str):
        item = None
        for key, value in self.dict.items():
            if isinstance(value, ModuleWrapper) and value.data.location == location:
                item = value

        return item

    def register_instrument(self, instrument: InstrumentContext):
        wrapper = InstrumentWrapper.from_instance(instrument, self)
        new_uuid = self.generate_new_id()
        self.dict[new_uuid] = wrapper
        return new_uuid

    def get_instrument_by_id(self, item_id: uuid.UUID):
        item = self.dict.get(item_id)

        if isinstance(item, InstrumentWrapper):
            return item

        return None

    def get_instrument_by_mount(self, mount: str):
        item = None
        for key, value in self.dict.items():
            if isinstance(value, InstrumentWrapper) and value.data.mount == mount:
                item = value

        return item

    def get_instrument_id_by_reference(self, instance: InstrumentContext):
        item = None
        for key, value in self.dict.items():
            if isinstance(value, InstrumentWrapper) and value.instrument_instance == instance:
                item = key

        return item

    def register_trash(self, trash: TrashBin | WasteChute, is_fixed: bool = False, is_waste_chute: bool = False, location: str = None):
        if is_fixed and isinstance(trash, TrashBin):
            self.has_fixed_trash = True
            self.dict[fixed_trash_id] = TrashBinWrapper(trash_bin_instance=trash, data=TrashBinData(deck_location=location, is_fixed=is_fixed))
            return fixed_trash_id

        if is_waste_chute and isinstance(trash, WasteChute):
            self.has_waste_chute = True
            self.dict[waste_chute_id] = trash
            return waste_chute_id

        if isinstance(trash, TrashBin):
            new_uuid = self.generate_new_id()
            self.dict[new_uuid] = TrashBinWrapper(trash_bin_instance=trash, data=TrashBinData(deck_location=location, is_fixed=False))
            return new_uuid

    def get_trash_by_id(self, item_id: uuid.UUID):
        item = self.dict.get(item_id)

        if isinstance(item, TrashBinWrapper):
            return item

        if isinstance(item, WasteChute):
            return item

        return None

    def get_trash_id_by_deck_location(self, location: str):
        item = None

        for key, value in self.dict.items():
            if isinstance(value, TrashBinWrapper) and value.data.deck_location == location:
                item = key

        return item

    def generate_new_id(self):
        return self.data.generate_new_id()


class DynamicItemsDict:
    dict: Dict[uuid.UUID, Well | Location] = {}
    data: Data

    def __init__(self, data: Data):
        self.data = data

    def get_well_location_by_id(self, item_id: uuid.UUID):
        return self.dict.get(item_id)

    def register_location(self, location: Location):
        for key, value in self.dict.items():
            if value == location:
                return key

        new_uuid = self.generate_new_id()
        self.dict[new_uuid] = location
        return new_uuid

    def unregister_locations(self, labware: Labware):
        for key, value in self.dict.items():
            if isinstance(value, Location) and value.labware == labware:
                self.dict.pop(key)

    def get_location_by_id(self, item_id: uuid.UUID):
        item = self.dict.get(item_id)

        if isinstance(item, Location):
            return item

        return None

    def register_wells_by_labware(self, labware: Labware):
        for key, value in self.dict.items():
            if value.parent == labware:
                return

        for well in labware.wells():
            new_uuid = self.generate_new_id()
            self.dict[new_uuid] = well

    def unregister_wells(self, labware: Labware):
        for well in labware.wells():
            if well.parent == labware:
                self.dict.pop(self.get_well_id_by_reference(well))

    def get_well_by_id(self, item_id: uuid.UUID):
        item = self.dict.get(item_id)

        if isinstance(item, Well):
            return item

        return None

    def get_well_id_by_reference(self, well: Well):
        item = None
        for key, value in self.dict.items():
            if value == well:
                return key
        return item

    def get_well_ids_by_reference(self, wells: List[Well]):
        items = []
        for well in wells:
            for key, value in self.dict.items():
                if value == well:
                    items.append(key)
        return items

    def generate_new_id(self):
        return self.data.generate_new_id()
